#define FASTJET_PACKAGE_STRING  "FastJet 3.4.0"
#define FASTJET_PACKAGE_VERSION  "3.4.0"
#define FASTJET_VERSION  "3.4.0"
#define FASTJET_VERSION_MAJOR       3
#define FASTJET_VERSION_MINOR       4
#define FASTJET_VERSION_PATCHLEVEL  0
#define FASTJET_VERSION_NUMBER      30400

/* The ATLASCone plugin is disabled by default*/
#undef FASTJET_ENABLE_PLUGIN_ATLASCONE 

/* The CDFJetClu and CDFMidPoint plugins are enabled by default*/
#define FASTJET_ENABLE_PLUGIN_CDFCONES 

/* The CMSIterativeCone plugin is disabled by default*/
#undef FASTJET_ENABLE_PLUGIN_CMSITERATIVECONE 

/* The D0RunICone plugin is disabled by default*/
#undef FASTJET_ENABLE_PLUGIN_D0RUNICONE 

/* The D0RunIICone plugin is disabled by default*/
#undef FASTJET_ENABLE_PLUGIN_D0RUNIICONE 

/* The EECambridge plugin is enabled by default*/
#define FASTJET_ENABLE_PLUGIN_EECAMBRIDGE 

/* The Jade plugin is enabled by default*/
#define FASTJET_ENABLE_PLUGIN_JADE 

/* The NestedDefs plugin is enabled by default*/
#define FASTJET_ENABLE_PLUGIN_NESTEDDEFS 

/* The PxCone plugin is disabled by default*/
#undef FASTJET_ENABLE_PLUGIN_PXCONE

/* The SISCone plugin is enabled by default*/
#define FASTJET_ENABLE_PLUGIN_SISCONE 

/* The TrackJet plugin is disabled by default*/
#undef FASTJET_ENABLE_PLUGIN_TRACKJET

/* The GridJet plugin is enabled by default */
#define FASTJET_ENABLE_PLUGIN_GRIDJET

/* end of plugin section */
